
#include "virtualcard.h"

#include "virtualsat.h"

/* Removed for lincs
std::ostream& operator<<(std::ostream& os, const VirtualCard& dt) {
    dt.print(os);
    return os;
}
*/  // Removed for lincs


VirtualCard::~VirtualCard() {

}
